#
# getCharts.pl
# ------------
#
# Automatic file transfer script
# ------------------------------
#
# Copyright (C) 2005-2018 Francis Breame
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# For details of the GNU General Public License see
# http://perldoc.perl.org/perlgpl.html, or write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
#
# Author:	Francis Breame G8ISI
# 		vf0123@btinternet.com
#
#
#
# 1.0.0	050220	Initial release.
#
# 1.1.0 050906	Add date variables to SITE, FILE, & FILE/file commands, as well as DIR & FILE/dir. Add %c variable.
#		Check that files have not already been xferred - can be overridden with OVERWRITE option.
#		Add option to shell to data manager (DATAMANAGER option added).
#
# 1.1.1 050911	Add %z date variable. Change OVERWRITE default to N. Minor fixes.
#
# 1.2.0 051127	Add %a, %b
#
# 1.3.0 051201	Add options to SITE : uid=/pw= (optional ftp & http authentication), dir=
#		Add SHUTDOWN command.
#		Change STAMP default to N.
#
# 1.3.1 060119	Fix to add leading zeroes to %a, %b
#
# 1.4.0 060205	Add filename globbing to ftp
#
# 1.5.0	060615	Add %%z, %%m, %%d, %%c, %%a, %%b (no leading zeroes)
#
# 1.6.0	060625	Add %n
#
# 1.7.0	061211	All times can now be specified as {h}h, {h}mm, or {h}h:mm.
#		This applies to repeat= as well, so that repeat times can include minutes and not
#		just specify whole hours.
#		Added STARTNOW global option to prevent transfer of files with times in the past on startup
#		Fixed bug with multiple : in urls.
#
# 1.8.0 080107	Add variables derived from the upload date/time of a source file (HTTP ONLY):
#		%fy,%fz,%fm,%fn,%fd,%fa,%fh,%fi,%fs.
#		Includes support for slot numbering (%fs).
#		Add SLOTSIZE option (global & file level).
#
# 1.8.1 080108	Add variable %fst (time-based slot).
#
# 1.8.2 080130	Fixed to allow = in urls. Any unrecognised keyword now has an implied FILE=.
#
# 1.9.0 080313	Mods suggested by Gerald Ihninger
#		Added STAMPFORMAT option (global & file level).
#		More accurate calculation of sleep time to next transfer slot.
#
# 1.10.0 081220	Add PLUGIN option (file level).
#
# 1.11.0 181103	Add https support.
#		Note that the Mozilla::CA file cacert.pem must be up to date.
#		Call Mozilla::CA::SSL_ca_file() (invoked by -c option) to find its location.
#		In my case
#			C:\Tools\perl_Strawberry_5.16.3.1_64\perl\vendor\lib\Mozilla\CA\cacert.pem
#		Enable use IO::Socket::SSL qw(debug3) for ssl details.
#

use Net::FTP;
use LWP::UserAgent;
use LWP::Protocol::https;
use Mozilla::CA;
#use IO::Socket::SSL qw(debug3);
use HTTP::request;
use HTTP::response;
use HTTP::headers;
use File::Path;
use File::Basename;
use File::Spec;
use Cwd;

use Text::Glob qw(match_glob);		# Text\glob.pm

use strict "refs";
use strict "vars";


####################################################################
my $ver		= "getCharts.pl v1.11.0";
####################################################################


my %q;		# Queue area - indexed by hhmm_nn
		# Subkeys are:
		#	tries		- try count
		#	sourcePath	- full path of source file, excluding file name
		#	sourceFile	- name of source file
		# 	destDir		- dir of dest file
		#	destFile	- name of dest file - if none use source file
		#	stamp		- timestamp dest file
		#	stampformat	- format for stamp
		#	slotsize	- size of timeslot when using slot naming
		#	manager		- use data manager for transfer
		#	uid		- authentication userid
		#	pw		- authentication password


# Global options - defaults are set in initDefaults
my $GstartT;
my $GrepT;
my $GdestDir;
my $Gstamp;
my $GstampFormat;
my $Gslotsize;
my $retryT;
my $maxTries;
my $debugLev;
my $inhibXfer;
my $ftppw;
my $utc;
my $overwrite;
my $datamanager;
my $shutdown;
my $startnow;

# Constants
my $ftpuid	= "anonymous";
my $dataFile	= "charts.txt";
my $dirSep	= "\\";		# Sep for dest filesystem. Source assumed to be / always

my @months = ('JAN','FEB','MAR','APR','MAY','JUN','JUL','AUG','SEP','OCT','NOV','DEC');

# Internal globals
my ($duffData,
   $duffArg,
   $dataFileArg,
   $curLine);



sub whichTime
{
	my $secs = shift;	# Time in secs
	return ($utc) ? (gmtime($secs)) : (localtime($secs));
}

sub getTimestamp
{
	# Returns yyyy/mm/dd hh:mm:ss - note local time not as set by UTC option

	my ($s,$mi,$h,$d,$m,$y);
	($s,$mi,$h,$d,$m,$y) = localtime(time);
	return sprintf "%04d/%02d/%02d %02d:%02d:%02d", $y+1900,$m+1,$d,$h,$mi,$s;
}

sub getTimestamp2
{
	# Returns formatted timestamp.
	# Formats are:
	# 0	_yyyymmdd_hhmmss
	# 1	_yyyymmddhhmm

	my @formats =
	(
		"_%04d%02d%02d_%02d%02d%02d",
		"_%04d%02d%02d%02d%02d"
	);
	my ($s,$mi,$h,$d,$m,$y);

	my $format = shift;
	$format=0 if (!defined $format || $format<0 || $format>$#formats);

	($s,$mi,$h,$d,$m,$y) = whichTime(time);
	return sprintf $formats[$format], $y+1900,$m+1,$d,$h,$mi,$s;
}

sub getNow
{
	# Returns current hhmm

	my (undef,$mi,$h) = whichTime(time);
	return sprintf "%02d%02d", $h,$mi;
}

sub getDate
{
	# Returns yyyymmdd

	my ($d,$m,$y);
	(undef,undef,undef,$d,$m,$y) = whichTime(time);
	return sprintf "%04d%02d%02d", $y+1900,$m+1,$d;
}

sub addHHMM
{
	# Add a number of mins to a time in form hhmm
	# If 3rd param defined, then result not reduced mod 24 hrs

	$_ = shift;
	/(^.*)(..$)/;
	my $t = ($1*60+$2+shift(@_));

	$t %= 1440 if (!defined shift(@_));

	return sprintf "%02d%02d", int($t/60), $t%60;
}

sub diffHHMM
{
	# Return difference in mins between two times in form hhmm, reduced mod 24 hrs

	$_ = shift;
	/(^.*)(..$)/;
	my $t1 = $1*60+$2;

	$_ = shift;
	/(^.*)(..$)/;
	my $t2 = $1*60+$2;

	return ($t2-$t1)%1440;
}

sub howLongSec
{
	# Return difference in secs between current time and given time in form hhmm,
	# reduced mod 24 hrs

	my ($s,$mi,$h,$t);

	$_ = shift;
	/(^.*)(..$)/;
	$t = ($1*60+$2)*60;

	($s,$mi,$h) = whichTime(time);

	return ($t-(($h*60+$mi)*60+$s))%86400;
}

sub diffTimes
{
	# Return difference between two times (in time struct) in form hh:mm:ss

	my $t1 = shift;
	my $t2 = shift;

	my $td = $t2 - $t1;

	return sprintf "%02d:%02d:%02d", int($td/3600), int($td/60)%60, $td%60;
}

sub time2mins
{
	# Convert time in form hhmm to minutes
	my $s = shift;
	return int($s/100)*60 + ($s%100);
}

sub isGlob
{
	# Check if url is ftp & contains a globbable filename

	my $url = shift;
	my $file = basename($url);
	return ($url=~/^ftp/i && $file=~/.*[[{*?]/);
}

sub mes
{
	my $s = shift;
	print getTimestamp.": $s\n";
}

sub debug
{
	my $s = shift;
	my $n = shift;
	$n = 1 if (!defined $n);

	print "$s\n"  if ($n<=$debugLev);
}

sub trim
{
	$_ = shift;
	s/^\s*//;
	s/\s*$//;
	return $_;
}

sub expandDate
{
	# Expand date variables in a path based on the current date
	# e.g. L:/charts/%y/%m/%d/xxx.gif
	#
	#	%y = yyyy - year
	#
	#	%z = yy - year
	#	%%z= as %z but no leading zeroes
	#
	#	%m = mm - month
	#	%%m= as %m but no leading zeroes
	#
	#	%n = mmm - month, 3 char text format
	#
	#	%d = dd - day in month
	#	%%d= as %d but no leading zeroes
	#
	# 	%c = dd - day in month, but for previous day. Forces other fields to decrement also
	#	%%c= as %c but no leading zeroes
	#
	#	%a = ddd - day in year
	#	%%a= as %a but no leading zeroes
	#
	#	%b = ddd - day in year, but for previous day as for %c
	#	%%b= as %b but no leading zeroes
	#
	#	%t = yyyy/mm/dd

	my $path = shift;

	my ($sS,$miS,$hS,$dS,$mS,$yS,$aS);
	my ($s,$mi,$h,$d,$dz,$m,$mz,$y,$z,$zz,$a,$az,$n,$t);
	my $secs = time;

	# Refer to yesterday if it includes %c %%c %b %%b
	$secs -= 86400 if ($path =~ /%c/i || $path =~ /%b/i);

	($sS,$miS,$hS,$dS,$mS,$yS,undef,$aS) = whichTime($secs);
	$yS += 1900;
	$mS++;
	$aS++;

	$y = sprintf "%04d", $yS;

	$z = sprintf "%02d", $yS % 100;
	$zz= sprintf "%d",   $yS % 100;

	$m = sprintf "%02s", $mS;
	$mz= sprintf "%s",   $mS;

	$n = $months[$mS-1];

	$d = sprintf "%02s", $dS;
	$dz= sprintf "%s",   $dS;

	$a = sprintf "%03s", $aS;
	$az= sprintf "%s",   $aS;

	$t = File::Spec->catfile($y,$m,$d);

	# Expand any variables

	$path =~ s/%y/$y/ig;

	$path =~ s/%%z/$zz/ig;
	$path =~ s/%z/$z/ig;

	$path =~ s/%%m/$mz/ig;
	$path =~ s/%m/$m/ig;

	$path =~ s/%n/$n/ig;

	$path =~ s/%%d/$dz/ig;
	$path =~ s/%d/$d/ig;

	$path =~ s/%%c/$dz/ig;
	$path =~ s/%c/$d/ig;

	$path =~ s/%%a/$az/ig;
	$path =~ s/%a/$a/ig;

	$path =~ s/%%b/$az/ig;
	$path =~ s/%b/$a/ig;

	$path =~ s/%t/$t/ig;

	return $path;
}

sub expandSourceDate
{
	# Expand date variables in a path based on a supplied time
	#
	#	%fy = yyyy - year
	#
	#	%fz = yy - year
	#	%%fz= as %fz but no leading zeroes
	#
	#	%fm = mm - month
	#	%%fm= as %fm but no leading zeroes
	#
	#	%fn = mmm - month, 3 char text format
	#
	#	%fd = dd - day in month
	#	%%fd= as %fd but no leading zeroes
	#
	#	%fa = ddd - day in year
	#	%%fa= as %fa but no leading zeroes
	#
	#	%fh = hh - hour
	#
	#	%fi = mm - min
	#
	#	%fs  = ss - slot number
	#	%%fs = as %fs but no leading zeroes
	#		e.g. if slotsize=30, then:
	#		23:45-00:14 = 0
	#		00:15-00:44 = 1
	#
	#	%fst  = hhmm - time-based slot number
	#

	my $path = shift;
	my $fileTime = shift;
	my $slotsize = shift;

	my ($oldPath,$i,$h,$d,$dz,$m,$mz,$y,$z,$zz,$a,$az,$sl,$slz,$slot,$tslot,$ret,$n);
	my ($sS,$miS,$hS,$dS,$mS,$yS,undef,$aS);

	$oldPath = $path;

	if (defined $fileTime && $fileTime!=0)
	{
		($sS,$miS,$hS,$dS,$mS,$yS,undef,$aS) = gmtime($fileTime);
		$yS += 1900;
		$mS++;
		$aS++;
	}
	else
	{
		$fileTime = $slot = $sS = $miS = $hS = $dS = $mS = $yS = $aS = 0;
	}

	$y = sprintf "%04d", $yS;

	$z = sprintf "%02d", $yS % 100;
	$zz= sprintf "%d",   $yS % 100;

	$m = sprintf "%02s", $mS;
	$mz= sprintf "%s",   $mS;

	$n = $months[$mS-1];

	$d = sprintf "%02s", $dS;
	$dz= sprintf "%s",   $dS;

	$a = sprintf "%03s", $aS;
	$az= sprintf "%s",   $aS;

	$i = sprintf "%02s", $miS;

	$h = sprintf "%02s", $hS;

	$slot = int(((($hS*60)+$miS+($slotsize/2))%1440)/$slotsize) if ($fileTime!=0);
	$sl = sprintf "%02d", $slot;
	$slz= sprintf "%d",   $slot;

	$tslot = $slot * $slotsize;
	$tslot = sprintf "%02d%02d", $tslot/60, $tslot%60;

	# Expand any variables

	$path =~ s/%fy/$y/ig;

	$path =~ s/%%fz/$zz/ig;
	$path =~ s/%fz/$z/ig;

	$path =~ s/%%fm/$mz/ig;
	$path =~ s/%fm/$m/ig;

	$path =~ s/%fn/$n/ig;

	$path =~ s/%%fd/$dz/ig;
	$path =~ s/%fd/$d/ig;

	$path =~ s/%%fa/$az/ig;
	$path =~ s/%fa/$a/ig;

	$path =~ s/%fi/$i/ig;

	$path =~ s/%fh/$h/ig;

	$path =~ s/%fst/$tslot/ig;

	$path =~ s/%%fs/$slz/ig;
	$path =~ s/%fs/$sl/ig;

	debug((sprintf "Time %02d:%02d:%02d mapped to slot %02d %s, size=%d", $hS,$miS,$sS,$slot,$tslot,$slotsize),1);

	if ($path eq $oldPath)	# Changed?
	{
		$ret = 0;
	}
	else
	{
		$ret = 1;
		mes "WARNING - no file upload time available. Slot number & time set to 0" if ($fileTime==0);
	}

	return ($path,$ret);
}


sub getFromPath
{
	# arg = %q index
	my $t = shift;

	my $fromDir  = $q{$t}{sourcePath};
	my $fromFile = $q{$t}{sourceFile};

	# Remove any trailing / from directory
	$fromDir =~ s+/$++;
	my $fromPath = $fromDir."/".$fromFile;

	# Expand any date variables
	return expandDate($fromPath);
}

sub getToPath
{
	# Form full dest pathname
	# e.g. L:/charts/%y/%m/%d/xxx{_stamp}.gif

	# arg = %q index
	my $t = shift;

	my $toDir  = $q{$t}{destDir};
	my $toFile = $q{$t}{destFile};

	# Remove any trailing \ from directory
	$toDir =~ s/$dirSep$//;

	# Expand any date variables
	$toDir  = expandDate($toDir);
	$toFile = expandDate($toFile);

	# Timestamp the filename before the . (dot)
	if ($q{$t}{stamp})
	{
		my $ts = getTimestamp2($q{$t}{stampformat});
		$toFile =~ s/\./$ts./;
	}

	return (length $toDir>0) ? File::Spec->catfile($toDir,$toFile) : $toFile;
}

sub checkDir
{
	# Ensure the directory exists, creating is necessary

	my $path = shift;

	my $dir = dirname($path);	# Must be a full path name including file

	my $ret = 1;

	if (!-d $dir)		# Create if necessary
	{
		debug ("Creating directory '$dir'",1);
		mkpath $dir;
		$ret=0 if (!-d $dir);
	}

	return $ret;
}

sub checkExists
{
	# Check if specified dest file already exists and is not to be overwritten

	my $toPath = shift;

	if (-e $toPath && !$overwrite)
	{
		mes "Destination file '$toPath' already exists and OVERWRITE option is not set";
		return 1;
	}
	else
	{
		return 0;
	}
}

sub doFTP
{
	my $url    = shift;
	my $toPath = shift;
	my $uid    = shift;
	my $pw     = shift;

	$uid = $ftpuid if (length $uid == 0);
	$pw  = $ftppw  if (length $pw  == 0);

	my ($ftp, $ret, $status, $fromPath, $fromDir, $fromFile, $toDir, $toFile, $host, $size, $file, @ls, @lsm);
	my ($n1, $n2, $ok);

	# Split into domain and source path
	$url =~ m|(.*://)(.*?)(/)(.*)|;
	$host = $2;
	$fromPath = $4;
	$fromDir  = dirname($fromPath);		# See File::Basename
	$fromFile = basename($fromPath);

	$toDir    = dirname($toPath);
	$toFile   = basename($toPath);

	mes "Getting from '$url' to '$toPath'";

	$ret = 0;				# Retry
	return 1 if ($inhibXfer);

	return 2 if (!checkDir($toPath));	# Ensure dest directories exist

	if ($ftp = Net::FTP->new($host))
	{
		mes "ftp connection to '$host' established";

	FTP: 	{
		$ftp->login($uid,$pw) or last FTP;
		mes "Logged in as $uid";

		$ftp->binary() or last FTP;
		$ftp->cwd($fromDir) or last FTP;

		if (isGlob($url))
		{
			# Globbed filename - possible multiple transfer

			@ls = ($ftp->ls) or last FTP;		# Get file list
			@lsm = match_glob($fromFile, @ls);	# Create globbed list
			$n1 = $#ls+1;
			$n2 = $#lsm+1;
			mes "Getting $n2 matched files of a total of $n1";
			last FTP if ($n2<=0);

			$ok = 0;

			foreach $file (sort @lsm)
			{
				# If globbing, dest filename always same as source
				$toPath = File::Spec->catfile($toDir,$file);

				if (!checkExists($toPath))
				{
					mes "Getting file '$file'";

					$size = $ftp->size($file) or last FTP;
					mes "File size = $size bytes";

					$ftp->get($file,$toPath) or last FTP;
					mes "Got file '$file'";
					$ok++;
				}
			}

			if ($ok==0)
			{
				$ret = 2;	# Don't retry (checkExists failed for all)
				last FTP;
			}
		}
		else
		{
			# Single file transfer
			mes "Getting file '$fromFile'";

			$size = $ftp->size($fromFile) or last FTP;
			mes "File size = $size bytes";

			$ftp->get($fromFile,$toPath) or last FTP;
			mes "Got file '$fromFile'";
		}

		$ret = 1;
		}	# end FTP block

		$status = $ftp->message;
		$ftp->quit();
	}
	else
	{
		$status = "ftp connection to '$host' failed - $@";
	}

	mes "FTP request to '$host' ".(($ret) ? "succeeded" : "failed");
	mes "Status = $status";

	return $ret;
}

sub doHTTP
{
	my $url    = shift;
	my $toPath = shift;
	my $uid    = shift;
	my $pw     = shift;
	my $slotsize = shift;

	my ($ua, $hdr, $request, $response, $data, $ret, $changed);

	$ret = 0;

	mes "Getting from '$url' to '$toPath'";

	$ua = LWP::UserAgent->new;

#	mes "SSL_ca_file = ".Mozilla::CA::SSL_ca_file();

	# Set https opts
	$ua->ssl_opts(
		SSL_ca_file => Mozilla::CA::SSL_ca_file(),
		verify_hostname => 1,
		protocols_allowed => ['http', 'https']);

	$hdr = HTTP::Headers->new;

	if (length $uid != 0)
	{
		$hdr->authorization_basic($uid, $pw);
		mes "Using userid '$uid'";
	}

	# Query the target
	# For the moment this is just to get any info needed for slot expansion
	$request = HTTP::Request->new(HEAD => $url, $hdr);

	return 1 if ($inhibXfer);

	$response = $ua->request($request);

	if ($response->is_success)
	{
		# Expand any slot variables
		($toPath, $changed) = expandSourceDate($toPath,$response->last_modified,$slotsize);
		#($toPath, $changed) = expandSourceDate($toPath,undef,$slotsize);

		if ($changed)
		{
			mes "File upload time is '".$response->header("last-modified")."'";
			mes "Destination changed to '$toPath'";

			return 2 if (checkExists($toPath));
		}

		return 2 if (!checkDir($toPath));	# Ensure dest directories exist

		$request->method(GET);
		$response = $ua->request($request);

		if ($response->is_success)
		{
  			if (open OUT, ">".$toPath)
			{
				binmode OUT;
				$data = $response->content_ref;
				print OUT $$data;
				$ret = 1;
			}
			else
			{
				mes "Can't open output file '$toPath'";
				$ret = 2;
			}
		}
	}

	mes "HTTP request to '$url' ".(($ret) ? "succeeded" : "failed");
	mes "Status = ".$response->status_line;
	mes "Header info:\n".$response->headers_as_string if ($debugLev>=1);

	close OUT if (defined OUT);

	return $ret;
}

sub doManager
{
	my $url    = shift;
	my $toPath = shift;

	my ($cmd, $cwd, $cdrive, $cpath);

	mes "Getting from '$url' to '$toPath' using data manager";

	# Needs full to-path, so if the name doesn't start with a drive, add the current one.
	# If it doesn't start with a \, add the whole current directory.
	if ($toPath !~ /^[A-Za-z]:/)
	{
		$cwd = getcwd();		# See Cwd module
		$cwd =~ s+/+$dirSep+g;

		$cwd =~ /(^[A-Za-z]:)(.*)/;	# Split into drive & path
		$cdrive = $1;
		$cpath = $2;

		$toPath = "$cpath$dirSep$toPath" if ($toPath !~ /^$dirSep/);
		$toPath = "$cdrive$toPath"
	}

	$cmd = $datamanager;			# Sub paths into the data manager command
	$cmd =~ s/%1/$url/;
	$cmd =~ s/%2/$toPath/;

	# Now shell out to $cmd and hope for the best

	# Note that getright expects a full pathname, including drive, for the dest.
	# It also desn't like invoking parallel transfers - hence have to use the /W option.

	debug ("Invoking data manager: $cmd",1);

	return 1 if ($inhibXfer);

	return 2 if (!checkDir($toPath));	# Ensure dest directories exist

	system $cmd;				# Execute manager program

	return ((-e $toPath) ? 1 : 0);
}

sub doPlugin
{
	my $t = shift;		# Index to %q

	my ($fromDir,$fromFile,$fromPath,$cmd);

	my $Q = '"';

	my $plugin = $q{$t}{plugin};

	# Assume that plugin is in same directory as this program
	$plugin = File::Spec->catfile(dirname($0),$plugin);	# $0 = us

	$fromDir  = $q{$t}{sourcePath};		# This is the base url
	$fromFile = $q{$t}{sourceFile};

	# Remove any trailing / from directory
	$fromDir =~ s+/$++;

	# Expand any date variables
	$fromFile = expandDate($fromFile);

	$cmd = "$plugin $Q$fromDir$Q $Q$fromFile$Q 2>&1";

	debug ("Calling plugin '$cmd'",1);
	mes "Calling plugin '$plugin'";

	$fromPath = `$cmd`;

	if ($? != 0)
	{
		mes "Plugin '$plugin' failed:";
		mes $fromPath;
		return undef;
	}

	return $fromPath;
}

sub doXfer
{
	# arg = %q index
	# Returns:	1 - ok
	#		0 - failed, retry
	#		2 - failed, don't retry

	my $t = shift;		# Index to %q
	my ($protocol);

	my $fromPath = getFromPath($t);
	my $toPath   = getToPath($t);
	my $manager  = $q{$t}{manager};
	my $uid      = $q{$t}{uid};
	my $pw       = $q{$t}{pw};
	my $slotsize = $q{$t}{slotsize};
	my $plugin   = $q{$t}{plugin};

	# Extract protocol
	$fromPath =~ /(.*?)(:.*)/;
	$protocol = lc $1;

	print "\n";

	# Check if file exists unless globbable, in which case the check has to be done during the multiple (ftp) transfer
	return 2 if (!isGlob($fromPath) && checkExists($toPath));

	# Call a plugin?
	if (defined $plugin)
	{
		return 0 if (!defined ($fromPath = doPlugin($t)));
	}

	# Do appropriate transfer
	if ($manager)
	{
		return doManager($fromPath, $toPath);
	}
	elsif ($protocol eq "http" || $protocol eq "https")
	{
		return doHTTP($fromPath, $toPath, $uid, $pw, $slotsize);
	}
	elsif ($protocol eq "ftp")
	{
		return doFTP($fromPath, $toPath, $uid, $pw);
	}
	else
	{
		mes "Unknown protocol '$fromPath'";
		return 2;
	}
}

sub parseErr
{
	my $s = shift;
	print STDERR "* Error in data file line $.: $s\n";
	print STDERR "     Line is '$curLine'\n";
	$duffData = 1;
}

sub argErr
{
	my $s = shift;
	print STDERR "* ERROR: invalid argument: $s\n";
	$duffArg = 1;
}

sub pres
{
	# Return true if present, i.e. defined and not null
	my $s = shift;
	return (defined $s && length $s>0);
}

sub chkPres
{
	my $name = shift;
	my $s = shift;

	if (!pres($s))
	{
		parseErr("Needs more information after '$name'");
		return 0;
	}
	else
	{
		return 1;
	}
}

sub chkNum
{
	my $name = shift;
	my $n = shift;
	my $n1 = shift;
	my $n2 = shift;

	if ($n !~ /^\d+$/ || $n<$n1 || $n>$n2)
	{
		parseErr("value of '$name' is '$n' - must be a number between $n1 and $n2");
		return 0;
	}
	else
	{
		return 1;
	}
}

sub chkTime
{
	# Allowed formats are {h}h, {h}hmm, {h}h:mm
	# Converts to hhmm, LH zero padded

	my $name = shift;
	my $n = shift;

	my $h = -1;
	my $m = -1;
	my $t;

	if ($n =~ /^\d{1,2}$/)
	{
		$h = $n;
		$m = 0;
	}
	elsif ($n =~ /^(\d{1,2})(\d\d)$/ || $n =~ /^(\d{1,2}):(\d\d)$/)
	{
		$h = $1;
		$m = $2;
	}

	#debug("n=$n h=$h m=$m",5);
	if ($h>=0 && $h<=23 && $m>=0 && $m<=59)
	{
		$t = sprintf "%04d", ($h*100 + $m);	# Zero pad
	#debug ("time=$t",5);
		return $t;
	}

	parseErr("value of '$name' is '$n' - must be a time in the form hh, hhmm, or hh:mm");
	return undef;
}

sub chkBool
{
	my $name = shift;
	my $s = shift;

	if (pres($s) && $s !~ /^(Y|N)/i)
	{
		parseErr("value of '$name' is '$s' - must be Y(es) or N(o)");
		return 0;
	}
	else
	{
		return 1;
	}
}

sub chkDataManager
{
	my $name = shift;
	my $s = shift;

	if (pres($s) && ($s !~ /%1/ || $s !~ /%2/))
	{
		parseErr("value of '$name' is '$s' - must contain both %1 and %2");
		return 0;
	}
	else
	{
		return 1;
	}
}

sub chkurl
{
	my $url = my $urlOrig = shift;
	my ($proto, $domain);

	$url = "http://".$url if ($url !~ m!.*://!);	# Default to http if omitted

	$url =~ m!(^\w+?)(://)(.*)!;
	$proto = $1;
	$domain = $3;
	if ($proto !~ /http|ftp/)
	{
		parseErr ("'$proto' is not a supported protocol in URL '$url'");
		return 0;
	}

	if (!defined $domain || ($domain !~ m!^[a-z][\w-]*?\.[\w\.-]*?(:.*|/.*|$)!i && $domain !~ m!^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}(:.*|/.*|$)!))
	{
		parseErr ("'$urlOrig' doesn't look much like a URL");
		return 0;
	}
	else
	{
		return 1;
	}
}

sub tourl
{
	my $url = shift;

	$url =~ s+/$++;			# Strip any trailing /
	return $url;
}

sub toBool
{
	my $s = shift;

	return (!pres($s) || ($s =~ /^Y/i)) ? 1 : 0;
}

sub duffOpt
{
	my $s = shift;
	parseErr ("unknown option '$s'");
}


sub initDefaults
{
	# Default global options
	$GstartT	= 100;
	$GrepT		= 0;
	$GdestDir	= "%t";
	$Gstamp		= 0;
	$GstampFormat	= 0;
	$Gslotsize	= 30;
	$retryT		= 30;
	$maxTries	= 2;
	$debugLev	= 0;
	$inhibXfer	= 0;
	$ftppw		= 'me@mymail.com';
	$utc    	= 1;
	$overwrite	= 0;
	$datamanager	= '"c:\program files\GetRight\getright.exe" /URL:"%1" /FILE:"%2" /O /W';
	$shutdown	= "";
	$startnow	= 0;
}

sub setupQ
{
	# Read the data file and setup the queue for all the file transfers in which we could possibly be interested

# Data file format
# ----------------
#
# #comment
#
# Defaults for local file settings. All of these + options can take #comment on the end.
# All keyword are case-insensitive.
#
# 	TIME=hhhh
# 	REPEAT=n
# 	DIR=L:/charts/%y/%m/%d
# 	STAMP=N
#	STAMPFORMAT=0
#	SLOTSIZE=n
#
# Global options
#
# 	RETRYT=n
# 	MAXTRIES=n
# 	DEBUGLEV=n
# 	INHIBXFER=N
# 	FTPPW=me@there.xx
# 	UTC=Y
# 	OVERWRITE=Y
# 	DATAMANAGER="i:\program files\GetRight\getright.exe" /URL:"%1" /FILE:"%2" /O
#	(%1=will be completed by full source path, %2 by full destination path)
#
# Site and file definitions:
#
# 	SITE=http://www.xx.yy#skip{=Y},u{id}=userid,p{w}=password#comment
#
# 	{FILE=}filename#skip{=Y},t{ime}=0010,r{epeat}=6,d{ir}=L:/charts/%y/%m/%d,f{ile}=xxx.gif,s{tamp}{=Y},m{anager}{=Y}#comment
#
# All directory & file names (DIR, SITE, FILE, FILE/dir, FILE/file) may include date variables, which are expanded as follows:
#	%y = yyyy
#	%z = yy
#	%m = mm
#	%d = dd
#	%c = dd but refers to previous day. If this is used then %d (if used) is also decremented,
#	     as are %m & %y if appropriate.
#	%t = yyyy\mm\dd
#
# Defaults for file options:
#	t{ime}		= global
#	r{epeat}	= global
#	d{ir}		= global
#	s{tamp}		= global
#	stampf{ormat}	= global
#	slot{size}	= global
#	m{anager}	= N
#	f{ile}		= incoming filename
#	plugin		= none
#			All filenames subject to stamp setting
#
# Globals:
#	Keyword		Var		Site/file override	Default
#       ---------------------------------------------------------------
#	time		$GstartT	F			0100
#	repeat		$GrepT		F			0
#	dir		$GdestDir	S+F			%t
#	stamp		$Gstamp		F			N
#	stampformat	$GstampFormat	F			0
#	slotsize	$Gslotsize	F			30
#	retryT		$retryT		N			30
#	maxTries	$maxTries	N			2
#	debugLev	$debugLev	N			0
#	inhibXfer	$inhibXfer	N			N
#	ftppw		$ftppw		N			me@mymail.com
#	utc		$utc		N			Y
#	overwrite	$overwrite	N			N
#	datamanager	$datamanager	N			c:\program files\GetRight\getright.exe /URL:"%1" /FILE:"%2" /O
#	shutdown	$shutdown	N			null
#	startnow	$startnow	N			N
#
# Booleans must be Y|N or null (=>Y)
#
# Stamp is 	_yyyymmdd_hhmmss 	- actual rx time
#


	my $fn = shift;

	my (%opts, @bits, @times, $kwField, $file, $fileSkip, $startT, $repT);
	my ($stamp, $stampFormat, $slotsize, $manager, $plugin, $destDir, $sourceFile, $destFile);
	my ($url, $siteSkip, $siteDir, $siteUID, $sitePW);	# Site-specific values
	my ($keyword, $kwVal, $optStr, $opt, $s, $t, $j, $jj, $key, $val, $now, $tmp);
	my %slots;	# Index hhmm - holds seq number for each minute slot

	initDefaults;
	$duffData = 0;
	$url = "";

	mes "Reading data file '$fn'";

	open f, $fn or die "Can't open $fn";

	while (<f>)
	{
		chomp;
		s/^ *//;	# trim
		s/ *$//;
		$curLine = $_;

		# This is where writing the code gets _monumentally_ BORING !!!!!

		# Line is: keyword=kw value#options#comments

		if (length>0 && !/^#/)				# Skip null and lines starting with #
		{

			debug ("\nLine $.:$_",3);

			/(.*?)(#|$)(.*?)(#|$)(.*)/;		# Split into (keyword=value)#(options)#(comments)

			$kwField = $1;				# keyword=value
			$optStr  = trim($3);

			$kwField =~ /(.*?)(=|$)(.*)/;		# Split into (keyword)=(value)
								# Note that value can include =
			$keyword = trim($1);
			$kwVal   = trim($3);

			@bits = split /,/, $optStr;		# Divide opts

			undef %opts;
			foreach $s (@bits)			# Place into hash
			{
				($key, $val) = split /=/, $s;
				$val = "" if (!defined $val);
				$opts{trim($key)} = trim($val);
			}

			debug ("Keyword=$keyword, value=$kwVal",3);
			debug (keys (%opts)." options found",3);

			$_ = $keyword;
			SWITCH: {
			/^SITE$/i && do 		# SITE=url[#opt]
			{
				$url = $siteUID = $sitePW = "";		# Site options default values
				$siteDir  = $GdestDir;
				$siteSkip = 0;

				$url = tourl($kwVal) if (chkurl($kwVal));

				# Parse opts
				foreach $j (keys %opts)
				{
					$val = $opts{$j};
					if ($j =~ /^skip$/i)
					{
						$siteSkip = toBool($val) if (chkBool($j,$val));
					}
					elsif ($j =~ /^d(ir$|$)/i)	# D{IR}
					{
						$siteDir = $val if (chkPres($j,$val));
					}
					elsif ($j =~ /^u(id$|$)/i)	# U{ID}
					{
						$siteUID = $val if (chkPres($j,$val));
					}
					elsif ($j =~ /^p(w$|$)/i)	# P{W}
					{
						$sitePW = $val if (chkPres($j,$val));
					}
					else
					{
						duffOpt($j);
					}
				}
				last SWITCH;
			};

			# For pure kw=value lines, ignore any options which may have been given


			/^TIME$/i && do 		# TIME=time
			{
				$GstartT = $tmp if (defined($tmp=chkTime($keyword,$kwVal)));
				last SWITCH;
			};


			/^REPEAT$/i && do 		# REPEAT=time
			{
				$GrepT = $tmp if (defined($tmp=chkTime($keyword,$kwVal)));
				last SWITCH;
			};

			/^MAXTRIES$/i && do 		# MAXTRIES=n
			{
				$maxTries = $kwVal if (chkNum($keyword,$kwVal,1,10));
				last SWITCH;
			};

			/^RETRYT$/i && do 		# RETRYT=n (mins)
			{
				$retryT = $kwVal if (chkNum($keyword,$kwVal,1,60));
				last SWITCH;
			};

			/^FTPPW$/i && do 		# FTPPW=string
			{
				$ftppw = $kwVal if (chkPres($keyword,$kwVal));
				last SWITCH;
			};

			/^DATAMANAGER$/i && do 		# DATAMANAGER=string
			{
				$datamanager = $kwVal if (chkDataManager($keyword,$kwVal));
				last SWITCH;
			};

			/^DIR$/i && do 			# DIR=string
			{
				$GdestDir = $kwVal if (chkPres($keyword,$kwVal));
				last SWITCH;
			};

			/^STAMP$/i && do 		# STAMP=Y|N
			{
				$Gstamp = toBool($kwVal) if (chkBool($keyword,$kwVal));
				last SWITCH;
			};

			/^STAMPFORMAT$/i && do 		# STAMPFORMAT=n
			{
				$GstampFormat = $kwVal if (chkNum($keyword,$kwVal,0,1));
				last SWITCH;
			};

			/^INHIBXFER$/i && do 		# INHIBXFER=Y|N
			{
				$inhibXfer = toBool($kwVal) if (chkBool($keyword,$kwVal));
				last SWITCH;
			};

			/^DEBUGLEV$/i && do 		# DEBUGLEV=n
			{
				$debugLev = $kwVal if (chkNum($keyword,$kwVal,0,100));
				last SWITCH;
			};

			/^UTC$/i && do 			# UTC=Y|N
			{
				$utc = toBool($kwVal) if (chkBool($keyword,$kwVal));
				last SWITCH;
			};

			/^OVERWRITE$/i && do 		# OVERWRITE=Y|N
			{
				$overwrite = toBool($kwVal) if (chkBool($keyword,$kwVal));
				last SWITCH;
			};

			/^SHUTDOWN$/i && do 		# SHUTDOWN=time
			{
				$shutdown = $tmp if (defined($tmp=chkTime($keyword,$kwVal)));
				last SWITCH;
			};

			/^STARTNOW$/i && do 		# STARTNOW=Y|N
			{
				$startnow = toBool($kwVal) if (chkBool($keyword,$kwVal));
				last SWITCH;
			};

			/^SLOTSIZE$/i && do 		# SLOTSIZE=n
			{
				$Gslotsize = $kwVal if (chkNum($keyword,$kwVal,15,120));
				last SWITCH;
			};

			# <<<<< MUST BE LAST CASE >>>>>
			# Anything else implies FILE=
			# The whole field is used even if it includes = which may occur in some urls
			if ($keyword !~ /^FILE$/i)
			{
				$kwVal = $kwField;
				$keyword = $_ = "FILE";
				debug ("Implied FILE=",3);
			}

			/^FILE$/i && do 		# FILE=file[#opts#comments]]
			{
#{FILE=}filename.gif#skip{=Y},t{ime}=0010,r{epeat}=6,d{ir}=L:/charts/%y/%m/%d,f{ile}=xxx.gif,s{tamp}{=Y}#comment

				if (length $url == 0)
				{
					parseErr("No site defined for this file");
				}

				last SWITCH if ($siteSkip);		# Skipping this site

				last SWITCH if (!chkPres($keyword,$kwVal));

				$fileSkip = 0;

				$sourceFile = $kwVal;
				$sourceFile =~ s+^/++;			# Strip any leading /
				$_          = "/".$sourceFile;		# Strip any leading directories
				m|(.*)(/)(.*$)|;
				$destFile   = $3;			# Default options to their global settings
				$startT     = $GstartT;
				$repT       = $GrepT;
				$destDir    = $siteDir;
				$stamp      = $Gstamp;
				$stampFormat= $GstampFormat;
				$slotsize   = $Gslotsize;
				$manager    = 0;
				undef $plugin;

				# Parse opts
				foreach $j (keys %opts)
				{
					$val = $opts{$j};
					debug("Processing opt $j=$val",3);

					if ($j =~ /^skip$/i)		# SKIP
					{
						$fileSkip = toBool($val) if (chkBool($j,$val));
					}
					elsif ($j =~ /^slot(size$|$)/i)	# SLOT{SIZE}
					{
						$slotsize = $val if (chkNum($j,$val,15,120));
					}
					elsif ($j =~ /^stampf(ormat$|$)/i)	# STAMPF{ORMAT}
					{
						$stampFormat = $val if (chkNum($j,$val,0,1));
					}
					elsif ($j =~ /^s(tamp$|$)/i)	# S{TAMP}
					{
						$stamp = toBool($val) if (chkBool($j,$val));
					}
					elsif ($j =~ /^t(ime$|$)/i)	# T{IME}
					{
						$startT = $tmp if (defined($tmp=chkTime($j,$val)));
					}
					elsif ($j =~ /^r(epeat$|$)/i)	# R{EPEAT}
					{
						$repT = $tmp if (defined($tmp=chkTime($j,$val)));
					}
					elsif ($j =~ /^d(ir$|$)/i)	# D{IR}
					{
						$destDir = $val if (chkPres($j,$val));
					}
					elsif ($j =~ /^m(anager$|$)/i)	# M{ANAGER}
					{
						$manager = toBool($val) if (chkBool($j,$val));
					}
					elsif ($j =~ /^f(ile$|$)/i)	# F{ILE}
					{
						$destFile = $val if (chkPres($j,$val));
						parseErr("cannot use the file= option to specify a destination file '$destFile'\n".
							"when wildcards are used in the source file name '$sourceFile'")
						if (isGlob("$url/$sourceFile") && basename($sourceFile) ne basename($destFile));

					}
					elsif ($j =~ /^plugin$/i)	# PLUGIN
					{
						$plugin = $val;
					}
					else
					{
						duffOpt($j);
					}
				}

				last SWITCH if ($fileSkip);		# Skip the file

				# At long last make queue entries for this file
				#%q subkeys:
				#	tries		- try count
				#	sourcePath	- full path of source file, excluding file name
				#	sourceFile	- name of source file
				# 	destDir		- dir of dest file
				#	destFile	- name of dest file - if none use source file
				#	stamp		- timestamp dest file
				#	slotsize	- timeslot naming size
				#	plugin		- plugin to translate source filename
				#	manager		- use data manager for transfer
				#	uid		- authentication userid
				#	pw		- authentication password


				# Create list of repeated events
				undef @times;
				debug ("Queueing from $startT by $repT",3);

				for ($t=$startT; $t<=2359; $t=addHHMM($t,time2mins($repT),"Y"))
				{
					push @times, sprintf "%04d",$t;

					last if ($repT<=0);
				}

				# .. and place them in the queue if necessary
				$now = getNow;
				for ($j=0; $j<=$#times; $j++)
				{
					$t = $times[$j];

					# If the time is in the past and there are future times,
					# skip the entry. Saves pulling in the same file several times
					# for past times if we start late in the day.

					next if ($t lt $now && $j < $#times);

					if (exists $slots{$t})
					{
						$slots{$t}++;		# Bump the seq no for the same slot time
					}
					else
					{
						$slots{$t} = "00";
					}

					$jj = $t."_".$slots{$t};	# Form q index including seq no
					$q{$jj}{tries}      = 0;
					$q{$jj}{sourcePath} = $url;
					$q{$jj}{sourceFile} = $sourceFile;
					$q{$jj}{destDir}    = $destDir;
					$q{$jj}{destFile}   = $destFile;
					$q{$jj}{stamp}      = $stamp;
					$q{$jj}{stampformat}= $stampFormat;
					$q{$jj}{slotsize}   = $slotsize;
					$q{$jj}{plugin}     = $plugin;
					$q{$jj}{manager}    = $manager;
					$q{$jj}{uid}        = $siteUID;
					$q{$jj}{pw}         = $sitePW;

					debug ("Slot $jj filled",3);
				}
				last SWITCH;
			};

			# Actually this can't now happen since anthing invalid implies FILE=
			#do 		# else ...
			#{
			#	parseErr ("Invalid keyword '$keyword'");
			#	last SWITCH;
			#};

			}	# End switch
		}
	}

	close f;

	mes ("Data read - global parameters are:\n".
		"time            = $GstartT\n".
		"repeat          = $GrepT\n".
		"dir             = $GdestDir\n".
		"stamp           = $Gstamp\n".
		"stampformat     = $GstampFormat\n".
		"slotsize        = $Gslotsize\n".
   		"retryT          = $retryT\n".
   		"maxTries        = $maxTries\n".
		"debugLev        = $debugLev\n".
		"inhibXfer       = $inhibXfer\n".
		"utc             = $utc\n".
   		"overwrite       = $overwrite\n".
   		"ftpPW           = $ftppw\n".
   		"datamanager     = $datamanager\n".
   		"shutdown        = $shutdown\n".
   		"startnow        = $startnow\n".
		"Q length        = ".(keys %q)."\n") if ($debugLev>=1);

	if ($duffData)
	{
		mes ("Quitting because of errors in data file '$fn'");
		exit;
	}
}

sub countQ
{
	# Determine number of q entries and number requiring retry
	my ($qsize, $t);
	my $retries = 0;

	$qsize = keys %q;
	foreach $t (keys %q)
	{
		$retries++ if($q{$t}{tries} > 0);
	}

	return ($qsize, $retries);
}

sub dumpQ
{
	my ($t);

	mes "------------------------- Q dump ---------------------------";
	foreach $t (sort keys %q)
	{
		print	"$t\tsource=$q{$t}{sourcePath}/$q{$t}{sourceFile}\n".
			"\tdest  =$q{$t}{destDir}$dirSep$q{$t}{destFile}\n".
			"\ttries =$q{$t}{tries}\tstamp=$q{$t}{stamp}\tstampformat=$q{$t}{stampformat}".
			"\tslotsize=$q{$t}{slotsize}\n".
			"\tmanager=$q{$t}{manager}\tuid   =$q{$t}{uid}\tpw=$q{$t}{pw}\n";
		print   "\tplugin=$q{$t}{plugin}\n" if (defined $q{$t}{plugin});
	}
	mes "------------------------------------------------------------";
}



#====================================================================================
# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> MAIN CODE <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
#====================================================================================




my ($qsize, $retries, $t, $tWaitMin, $tWaitSec, $nextSlot, $nextSlotT, $now, $day, $lastDay, $startedAt);
my ($ret, $xstart, $xstop, $nextT, $nextTnomod24);
my ($newSTDOUT, $newSTDERR);


# Process arguments
#	getCharts.pl config file [redirect STDOUT [redirect STDERR]]
# or
#	getCharts.pl -c ... to print the location of the ssl certificates file if replacing it

$duffArg = 0;

$dataFileArg = shift;

if (defined $dataFileArg && $dataFileArg =~ /^-c/i)
{
	print "SSL_ca_file location = ".Mozilla::CA::SSL_ca_file()."\n";
	exit;
}

$dataFile = $dataFileArg if (defined $dataFileArg);
argErr ("Non-existent data file '$dataFile'") if (!-f $dataFile);

exit if ($duffArg);

$newSTDOUT = shift;
open STDOUT, ">>".$newSTDOUT or die "Can't open $newSTDOUT" if (defined $newSTDOUT);

$newSTDERR = shift;
open STDERR, ">>".$newSTDERR or die "Can't open $newSTDERR" if (defined $newSTDERR);



mes "";
mes "===========================================================================================";
mes "Starting $ver running under account '".getlogin."'";

setupQ ($dataFile);
dumpQ;

$lastDay   = getDate;		# Current day
$startedAt = getNow;		# Time of intital start


TRANSFERCYCLE: while (1)
{
	# On wake up

	$now = getNow;
	$day = getDate;

	if ($day != $lastDay)
	{
		# Gone past midnight so reset the queue for a new day
		setupQ ($dataFile);
		dumpQ if ($debugLev >= 1);

		$startnow = 0;		# Switch off if set, now that we have started a new day
	}

	$lastDay = $day;


	# Scan the queue looking for events to process

	QSCAN: foreach $t (sort keys %q)
	{
		$nextSlotT = substr($t,0,4);			# Time of the next slot

		last QSCAN if ($nextSlotT gt $now);		# All the rest are in the future
								# so finished for now

		if ($startnow && ($nextSlotT lt $startedAt))	# Skip past entries if $startnow is set
		{
			delete $q{$t};
			next QSCAN;
		}

		# Attempt to process the event
		$q{$t}{tries}++;
		$xstart = time;
		$ret = doXfer($t);
		$xstop = time;

		# Handle the event's outcome
		if ($ret==0 || $ret==2)
		{
			# Unsuccesful - 0=retry, 2=don't
			mes "Transfer of '".getFromPath($t)."' slot $t failed";

			if ($ret==0)
			{
				if ($q{$t}{tries} >= $maxTries)
				{
					mes "Giving up after $q{$t}{tries} tries";
					delete $q{$t};
				}
				else
				{
					mes "Will retry later";
				}
			}
			elsif ($ret==2)
			{
				mes "Won't retry";
				delete $q{$t};
			}
		}
		else
		{
			# Success!
			mes "Transfer of '".getFromPath($t)."' slot $t succeeded in ".diffTimes($xstart,$xstop);
			delete $q{$t};
		}
	}
	dumpQ if ($debugLev >= 2);

	($qsize, $retries) = countQ;

	# If there are retries outstanding, wait till the next retryT period.
	# If not, wait till the next scheduled transfer. If none, wait till midnight.

	($nextSlot) = (sort keys %q);
	$nextSlot = addHHMM($now, $retryT) if ($retries>0);
	$nextSlot = "0000_00" if (!defined $nextSlot);	# i.e. next midnight
	$now = getNow;
	$tWaitSec = howLongSec(substr($nextSlot,0,4));
	$tWaitMin = int($tWaitSec/60+0.99);

	print "\n";

	$nextT        = addHHMM($now, $tWaitMin);
	$nextTnomod24 = addHHMM($now, $tWaitMin, "Y");

	mes ("End of transfer cycle - will start next cycle in $tWaitMin min ($tWaitSec sec) at $nextT");
	mes ("$qsize transfer".(($qsize==1)?"":"s")." remaining, of which $retries ".
		(($retries==1)?"is a retry":"are retries"));

	# Finish if shutdown has been specified and the next transfer will start after it
	if (length $shutdown!=0 && $shutdown <= $nextTnomod24)
	{
		mes ("Quitting since this is after the shutdown time of $shutdown");
		last TRANSFERCYCLE;
	}

	sleep $tWaitSec;
}


#
# NB We can install signal handlers like the following.
# There is a problem in that if the signal occurs during the sleep, it is not caught until the sleep ends.
# Fix sometime.
#
#sub catchSig
#{
#	my $signame = shift;
#	die "Exiting due to signal $signame - probably killed by user -";
#}
#
#
# Install signal handlers
#$SIG{INT} = \&catchSig;
#$SIG{HUP} = \&catchSig;
#$SIG{TERM} = \&catchSig;
#$SIG{STOP} = \&catchSig;
#$SIG{KILL} = \&catchSig;
#

